/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.ext.rbvi.chimera;

import jalview.api.AlignmentViewPanel;
import jalview.api.FeatureRenderer;
import jalview.api.SequenceRenderer;
import jalview.bin.Cache;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.ColumnSelection;
import jalview.datamodel.PDBEntry;
import jalview.datamodel.SequenceI;
import jalview.schemes.ColourSchemeI;
import jalview.schemes.ResidueProperties;
import jalview.structure.StructureMapping;
import jalview.structure.StructureMappingcommandSet;
import jalview.structure.StructureSelectionManager;
import jalview.structures.models.AAStructureBindingModel;
import jalview.util.Comparison;
import jalview.util.MessageManager;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import ext.edu.ucsf.rbvi.strucviz2.ChimeraManager;
import ext.edu.ucsf.rbvi.strucviz2.ChimeraModel;
import ext.edu.ucsf.rbvi.strucviz2.StructureManager;
import ext.edu.ucsf.rbvi.strucviz2.StructureManager.ModelType;

public abstract class JalviewChimeraBinding extends AAStructureBindingModel
{

  private static final boolean debug = false;

  private static final String PHOSPHORUS = "P";

  private static final String ALPHACARBON = "CA";

  private StructureManager csm;

  private ChimeraManager viewer;

  /*
   * set if chimera state is being restored from some source - instructs binding
   * not to apply default display style when structure set is updated for first
   * time.
   */
  private boolean loadingFromArchive = false;

  /*
   * flag to indicate if the Chimera viewer should ignore sequence colouring
   * events from the structure manager because the GUI is still setting up
   */
  private boolean loadingFinished = true;

  /*
   * state flag used to check if the Chimera viewer's paint method can be called
   */
  private boolean finishedInit = false;

  private List<String> atomsPicked = new ArrayList<String>();

  private List<String> chainNames;

  private Map<String, String> chainFile;

  public String fileLoadingError;

  /*
   * Map of ChimeraModel objects keyed by PDB full local file name
   */
  private Map<String, List<ChimeraModel>> chimeraMaps = new LinkedHashMap<String, List<ChimeraModel>>();

  /*
   * the default or current model displayed if the model cannot be identified
   * from the selection message
   */
  private int frameNo = 0;

  private String lastCommand;

  private String lastMessage;

  private boolean loadedInline;

  /**
   * current set of model filenames loaded
   */
  String[] modelFileNames = null;

  String lastMousedOverAtomSpec;

  private List<String> lastReply;

  /**
   * Open a PDB structure file in Chimera and set up mappings from Jalview.
   * 
   * We check if the PDB model id is already loaded in Chimera, if so don't
   * reopen it. This is the case if Chimera has opened a saved session file.
   * 
   * @param pe
   * @return
   */
  public boolean openFile(PDBEntry pe)
  {
    String file = pe.getFile();
    try
    {
      List<ChimeraModel> modelsToMap = new ArrayList<ChimeraModel>();
      List<ChimeraModel> oldList = viewer.getModelList();
      boolean alreadyOpen = false;

      /*
       * If Chimera already has this model, don't reopen it, but do remap it.
       */
      for (ChimeraModel open : oldList)
      {
        if (open.getModelName().equals(pe.getId()))
        {
          alreadyOpen = true;
          modelsToMap.add(open);
        }
      }

      /*
       * If Chimera doesn't yet have this model, ask it to open it, and retrieve
       * the model names added by Chimera.
       */
      if (!alreadyOpen)
      {
        viewer.openModel(file, pe.getId(), ModelType.PDB_MODEL);
        modelsToMap = viewer.getModelList();
        modelsToMap.removeAll(oldList);
      }

      chimeraMaps.put(file, modelsToMap);

      if (getSsm() != null)
      {
        getSsm().addStructureViewerListener(this);
        // ssm.addSelectionListener(this);
        FeatureRenderer fr = getFeatureRenderer(null);
        if (fr != null)
        {
          fr.featuresAdded();
        }
        refreshGUI();
      }
      return true;
    } catch (Exception q)
    {
      log("Exception when trying to open model " + file + "\n"
              + q.toString());
      q.printStackTrace();
    }
    return false;
  }

  /**
   * Constructor
   * 
   * @param ssm
   * @param pdbentry
   * @param sequenceIs
   * @param chains
   * @param protocol
   */
  public JalviewChimeraBinding(StructureSelectionManager ssm,
          PDBEntry[] pdbentry, SequenceI[][] sequenceIs, String[][] chains,
          String protocol)
  {
    super(ssm, pdbentry, sequenceIs, chains, protocol);
    viewer = new ChimeraManager(
            csm = new ext.edu.ucsf.rbvi.strucviz2.StructureManager(true));
  }

  /**
   * Constructor
   * 
   * @param ssm
   * @param theViewer
   */
  public JalviewChimeraBinding(StructureSelectionManager ssm,
          ChimeraManager theViewer)
  {
    super(ssm, null);
    viewer = theViewer;
    csm = viewer.getStructureManager();
  }

  /**
   * Construct a title string for the viewer window based on the data Jalview
   * knows about
   * 
   * @param verbose
   * @return
   */
  public String getViewerTitle(boolean verbose)
  {
    return getViewerTitle("Chimera", verbose);
  }

  /**
   * prepare the view for a given set of models/chains. chainList contains
   * strings of the form 'pdbfilename:Chaincode'
   * 
   * @param toshow
   *          list of chains to make visible
   */
  public void centerViewer(List<String> toshow)
  {
    StringBuilder cmd = new StringBuilder(64);
    int mlength, p;
    for (String lbl : toshow)
    {
      mlength = 0;
      do
      {
        p = mlength;
        mlength = lbl.indexOf(":", p);
      } while (p < mlength && mlength < (lbl.length() - 2));
      // TODO: lookup each pdb id and recover proper model number for it.
      cmd.append("#" + getModelNum(chainFile.get(lbl)) + "."
              + lbl.substring(mlength + 1) + " or ");
    }
    if (cmd.length() > 0)
    {
      cmd.setLength(cmd.length() - 4);
    }
    String cmdstring = cmd.toString();
    evalStateCommand("~display #*; ~ribbon #*; ribbon " + cmdstring
            + ";focus " + cmdstring, false);
  }

  /**
   * Close down the Jalview viewer, and (optionally) the associated Chimera
   * window.
   */
  public void closeViewer(boolean closeChimera)
  {
    getSsm().removeStructureViewerListener(this, this.getPdbFile());
    if (closeChimera)
    {
      viewer.exitChimera();
    }
    lastCommand = null;
    viewer = null;
    releaseUIResources();
  }

  public void colourByChain()
  {
    colourBySequence = false;
    evalStateCommand("rainbow chain", false);
  }

  public void colourByCharge()
  {
    colourBySequence = false;
    evalStateCommand(
            "color white;color red ::ASP;color red ::GLU;color blue ::LYS;color blue ::ARG;color yellow ::CYS",
            false);
  }

  /**
   * superpose the structures associated with sequences in the alignment
   * according to their corresponding positions.
   */
  public void superposeStructures(AlignmentI alignment)
  {
    superposeStructures(alignment, -1, null);
  }

  /**
   * superpose the structures associated with sequences in the alignment
   * according to their corresponding positions. ded)
   * 
   * @param refStructure
   *          - select which pdb file to use as reference (default is -1 - the
   *          first structure in the alignment)
   */
  public void superposeStructures(AlignmentI alignment, int refStructure)
  {
    superposeStructures(alignment, refStructure, null);
  }

  /**
   * superpose the structures associated with sequences in the alignment
   * according to their corresponding positions. ded)
   * 
   * @param refStructure
   *          - select which pdb file to use as reference (default is -1 - the
   *          first structure in the alignment)
   * @param hiddenCols
   *          TODO
   */
  public void superposeStructures(AlignmentI alignment, int refStructure,
          ColumnSelection hiddenCols)
  {
    superposeStructures(new AlignmentI[]
    { alignment }, new int[]
    { refStructure }, new ColumnSelection[]
    { hiddenCols });
  }

  public void superposeStructures(AlignmentI[] _alignment,
          int[] _refStructure, ColumnSelection[] _hiddenCols)
  {
    assert (_alignment.length == _refStructure.length && _alignment.length != _hiddenCols.length);
    StringBuilder allComs = new StringBuilder(128); // Chimera superposition cmd
    String[] files = getPdbFile();
    // check to see if we are still waiting for Chimera files
    long starttime = System.currentTimeMillis();
    boolean waiting = true;
    do
    {
      waiting = false;
      for (String file : files)
      {
        try
        {
          // HACK - in Jalview 2.8 this call may not be threadsafe so we catch
          // every possible exception
          StructureMapping[] sm = getSsm().getMapping(file);
          if (sm == null || sm.length == 0)
          {
            waiting = true;
          }
        } catch (Exception x)
        {
          waiting = true;
        } catch (Error q)
        {
          waiting = true;
        }
      }
      // we wait around for a reasonable time before we give up
    } while (waiting
            && System.currentTimeMillis() < (10000 + 1000 * files.length + starttime));
    if (waiting)
    {
      System.err
              .println("RUNTIME PROBLEM: Chimera seems to be taking a long time to process all the structures.");
      return;
    }
    refreshPdbEntries();
    StringBuffer selectioncom = new StringBuffer();
    for (int a = 0; a < _alignment.length; a++)
    {
      int refStructure = _refStructure[a];
      AlignmentI alignment = _alignment[a];
      ColumnSelection hiddenCols = _hiddenCols[a];
      if (a > 0
              && selectioncom.length() > 0
              && !selectioncom.substring(selectioncom.length() - 1).equals(
                      " "))
      {
        selectioncom.append(" ");
      }
      // process this alignment
      if (refStructure >= files.length)
      {
        System.err.println("Invalid reference structure value "
                + refStructure);
        refStructure = -1;
      }
      if (refStructure < -1)
      {
        refStructure = -1;
      }

      boolean matched[] = new boolean[alignment.getWidth()];
      for (int m = 0; m < matched.length; m++)
      {

        matched[m] = (hiddenCols != null) ? hiddenCols.isVisible(m) : true;
      }

      int commonrpositions[][] = new int[files.length][alignment.getWidth()];
      String isel[] = new String[files.length];
      String[] targetC = new String[files.length];
      String[] chainNames = new String[files.length];
      String[] atomSpec = new String[files.length];
      for (int pdbfnum = 0; pdbfnum < files.length; pdbfnum++)
      {
        StructureMapping[] mapping = getSsm().getMapping(files[pdbfnum]);
        // RACE CONDITION - getMapping only returns Jmol loaded filenames once
        // Jmol callback has completed.
        if (mapping == null || mapping.length < 1)
        {
          throw new Error(MessageManager.getString("error.implementation_error_chimera_getting_data"));
        }
        int lastPos = -1;
        final int seqCountForPdbFile = getSequence()[pdbfnum].length;
        for (int s = 0; s < seqCountForPdbFile; s++)
        {
          for (int sp, m = 0; m < mapping.length; m++)
          {
            final SequenceI theSequence = getSequence()[pdbfnum][s];
            if (mapping[m].getSequence() == theSequence
                    && (sp = alignment.findIndex(theSequence)) > -1)
            {
              if (refStructure == -1)
              {
                refStructure = pdbfnum;
              }
              SequenceI asp = alignment.getSequenceAt(sp);
              for (int r = 0; r < matched.length; r++)
              {
                if (!matched[r])
                {
                  continue;
                }
                matched[r] = false; // assume this is not a good site
                if (r >= asp.getLength())
                {
                  continue;
                }

                if (Comparison.isGap(asp.getCharAt(r)))
                {
                  // no mapping to gaps in sequence
                  continue;
                }
                int t = asp.findPosition(r); // sequence position
                int apos = mapping[m].getAtomNum(t);
                int pos = mapping[m].getPDBResNum(t);

                if (pos < 1 || pos == lastPos)
                {
                  // can't align unmapped sequence
                  continue;
                }
                matched[r] = true; // this is a good ite
                lastPos = pos;
                // just record this residue position
                commonrpositions[pdbfnum][r] = pos;
              }
              // create model selection suffix
              isel[pdbfnum] = "#" + pdbfnum;
              if (mapping[m].getChain() == null
                      || mapping[m].getChain().trim().length() == 0)
              {
                targetC[pdbfnum] = "";
              }
              else
              {
                targetC[pdbfnum] = "." + mapping[m].getChain();
              }
              chainNames[pdbfnum] = mapping[m].getPdbId()
                      + targetC[pdbfnum];
              atomSpec[pdbfnum] = asp.getRNA() != null ? PHOSPHORUS : ALPHACARBON;
              // move on to next pdb file
              s = seqCountForPdbFile;
              break;
            }
          }
        }
      }

      // TODO: consider bailing if nmatched less than 4 because superposition
      // not
      // well defined.
      // TODO: refactor superposable position search (above) from jmol selection
      // construction (below)

      String[] selcom = new String[files.length];
      int nmatched = 0;
      String sep = "";
      // generate select statements to select regions to superimpose structures
      {
        for (int pdbfnum = 0; pdbfnum < files.length; pdbfnum++)
        {
          String chainCd = targetC[pdbfnum];
          int lpos = -1;
          boolean run = false;
          StringBuffer molsel = new StringBuffer();
          for (int r = 0; r < matched.length; r++)
          {
            if (matched[r])
            {
              if (pdbfnum == 0)
              {
                nmatched++;
              }
              if (lpos != commonrpositions[pdbfnum][r] - 1)
              {
                // discontinuity
                if (lpos != -1)
                {
                  molsel.append((run ? "" : ":") + lpos);
                  molsel.append(chainCd);
                  molsel.append(",");
                }
              }
              else
              {
                // continuous run - and lpos >-1
                if (!run)
                {
                  // at the beginning, so add dash
                  molsel.append(":" + lpos);
                  molsel.append("-");
                }
                run = true;
              }
              lpos = commonrpositions[pdbfnum][r];
              // molsel.append(lpos);
            }
          }
          // add final selection phrase
          if (lpos != -1)
          {
            molsel.append((run ? "" : ":") + lpos);
            molsel.append(chainCd);
            // molsel.append("");
          }
          if (molsel.length() > 1)
          {
            selcom[pdbfnum] = molsel.toString();
            selectioncom.append("#" + pdbfnum);
            selectioncom.append(selcom[pdbfnum]);
            selectioncom.append(" ");
            if (pdbfnum < files.length - 1)
            {
              selectioncom.append("| ");
            }
          }
          else
          {
            selcom[pdbfnum] = null;
          }
        }
      }
      StringBuilder command = new StringBuilder(256);
      for (int pdbfnum = 0; pdbfnum < files.length; pdbfnum++)
      {
        if (pdbfnum == refStructure || selcom[pdbfnum] == null
                || selcom[refStructure] == null)
        {
          continue;
        }
        if (command.length() > 0)
        {
          command.append(";");
        }

        /*
         * Form Chimera match command, from the 'new' structure to the
         * 'reference' structure e.g. (residues 1-91, chain B/A, alphacarbons):
         * 
         * match #1:1-91.B@CA #0:1-91.A@CA
         * 
         * @see
         * https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/midas/match.html
         */
        command.append("match #" + pdbfnum /* +".1" */);
        // TODO: handle sub-models
        command.append(selcom[pdbfnum]);
        command.append("@" + atomSpec[pdbfnum]);
        command.append(" #" + refStructure /* +".1" */);
        command.append(selcom[refStructure]);
        command.append("@" + atomSpec[refStructure]);
      }
      if (selectioncom.length() > 0)
      {
        if (debug)
        {
          System.out.println("Select regions:\n" + selectioncom.toString());
          System.out.println("Superimpose command(s):\n"
                  + command.toString());
        }
        allComs.append("~display all; chain @CA|P; ribbon "
                + selectioncom.toString() + ";"+command.toString());
        // selcom.append("; ribbons; ");
      }
    }
    if (selectioncom.length() > 0)
    {// finally, mark all regions that were superposed.
      if (selectioncom.substring(selectioncom.length() - 1).equals("|"))
      {
        selectioncom.setLength(selectioncom.length() - 1);
      }
      if (debug)
      {
        System.out.println("Select regions:\n" + selectioncom.toString());
      }
      allComs.append("; ~display all; chain @CA|P; ribbon "
              + selectioncom.toString() + "; focus");
      // evalStateCommand("select *; backbone; select "+selcom.toString()+"; cartoons; center "+selcom.toString());
      evalStateCommand(allComs.toString(), true /* false */);
    }
    
  }

  private void checkLaunched()
  {
    if (!viewer.isChimeraLaunched())
    {
      viewer.launchChimera(StructureManager.getChimeraPaths());
    }
    if (!viewer.isChimeraLaunched())
    {
      log("Failed to launch Chimera!");
    }
  }

  /**
   * Answers true if the Chimera process is still running, false if ended or not
   * started.
   * 
   * @return
   */
  public boolean isChimeraRunning()
  {
    return viewer.isChimeraLaunched();
  }

  /**
   * Send a command to Chimera, launching it first if necessary, and optionally
   * log any responses.
   * 
   * @param command
   * @param logResponse
   */
  public void evalStateCommand(final String command, boolean logResponse)
  {
    viewerCommandHistory(false);
    checkLaunched();
    if (lastCommand == null || !lastCommand.equals(command))
    {
      // trim command or it may never find a match in the replyLog!!
      lastReply = viewer.sendChimeraCommand(command.trim(), logResponse);
      if (debug && logResponse)
      {
        log("Response from command ('" + command + "') was:\n" + lastReply);
      }
    }
    viewerCommandHistory(true);
    lastCommand = command;
  }

  /**
   * colour any structures associated with sequences in the given alignment
   * using the getFeatureRenderer() and getSequenceRenderer() renderers but only
   * if colourBySequence is enabled.
   */
  public void colourBySequence(boolean showFeatures,
          jalview.api.AlignmentViewPanel alignmentv)
  {
    if (!colourBySequence || !loadingFinished)
    {
      return;
    }
    if (getSsm() == null)
    {
      return;
    }
    String[] files = getPdbFile();

    SequenceRenderer sr = getSequenceRenderer(alignmentv);

    FeatureRenderer fr = null;
    if (showFeatures)
    {
      fr = getFeatureRenderer(alignmentv);
    }
    AlignmentI alignment = alignmentv.getAlignment();

    for (jalview.structure.StructureMappingcommandSet cpdbbyseq : getColourBySequenceCommands(
            files, sr, fr, alignment))
    {
      for (String command : cpdbbyseq.commands)
      {
        executeWhenReady(command);
      }
    }
  }

  /**
   * @param files
   * @param sr
   * @param fr
   * @param alignment
   * @return
   */
  protected StructureMappingcommandSet[] getColourBySequenceCommands(
          String[] files, SequenceRenderer sr, FeatureRenderer fr,
          AlignmentI alignment)
  {
    return ChimeraCommands.getColourBySequenceCommand(getSsm(), files,
            getSequence(), sr, fr, alignment);
  }

  /**
   * @param command
   */
  protected void executeWhenReady(String command)
  {
    waitForChimera();
    evalStateCommand(command, false);
    waitForChimera();
  }

  private void waitForChimera()
  {
    while (viewer != null && viewer.isBusy())
    {
      try {
        Thread.sleep(15);
      } catch (InterruptedException q)
      {}
    }
  }



  // End StructureListener
  // //////////////////////////

  public Color getColour(int atomIndex, int pdbResNum, String chain,
          String pdbfile)
  {
    if (getModelNum(pdbfile) < 0)
    {
      return null;
    }
    log("get model / residue colour attribute unimplemented");
    return null;
  }

  /**
   * returns the current featureRenderer that should be used to colour the
   * structures
   * 
   * @param alignment
   * 
   * @return
   */
  public abstract FeatureRenderer getFeatureRenderer(
          AlignmentViewPanel alignment);

  /**
   * instruct the Jalview binding to update the pdbentries vector if necessary
   * prior to matching the jmol view's contents to the list of structure files
   * Jalview knows about.
   */
  public abstract void refreshPdbEntries();

  private int getModelNum(String modelFileName)
  {
    String[] mfn = getPdbFile();
    if (mfn == null)
    {
      return -1;
    }
    for (int i = 0; i < mfn.length; i++)
    {
      if (mfn[i].equalsIgnoreCase(modelFileName))
      {
        return i;
      }
    }
    return -1;
  }

  /**
   * map between index of model filename returned from getPdbFile and the first
   * index of models from this file in the viewer. Note - this is not trimmed -
   * use getPdbFile to get number of unique models.
   */
  private int _modelFileNameMap[];

  // ////////////////////////////////
  // /StructureListener
  public synchronized String[] getPdbFile()
  {
    if (viewer == null)
    {
      return new String[0];
    }
    // if (modelFileNames == null)
    // {
    // Collection<ChimeraModel> chimodels = viewer.getChimeraModels();
    // _modelFileNameMap = new int[chimodels.size()];
    // int j = 0;
    // for (ChimeraModel chimodel : chimodels)
    // {
    // String mdlName = chimodel.getModelName();
    // }
    // modelFileNames = new String[j];
    // // System.arraycopy(mset, 0, modelFileNames, 0, j);
    // }

    return chimeraMaps.keySet().toArray(
            modelFileNames = new String[chimeraMaps.size()]);
  }

  /**
   * map from string to applet
   */
  public Map getRegistryInfo()
  {
    // TODO Auto-generated method stub
    return null;
  }

  /**
   * returns the current sequenceRenderer that should be used to colour the
   * structures
   * 
   * @param alignment
   * 
   * @return
   */
  public abstract SequenceRenderer getSequenceRenderer(
          AlignmentViewPanel alignment);

  /**
   * Construct and send a command to highlight an atom.
   * 
   * <pre>
   * Done by generating a command like (to 'highlight' position 44)
   *   ~select #0:43.C;select #0:44.C
   * Note this removes the selection from the previous position.
   * </pre>
   */
  public void highlightAtom(int atomIndex, int pdbResNum, String chain,
          String pdbfile)
  {
    List<ChimeraModel> cms = chimeraMaps.get(pdbfile);
    if (cms != null)
    {
      StringBuilder sb = new StringBuilder();
      sb.append(" #" + cms.get(0).getModelNumber());
      sb.append(":" + pdbResNum);
      if (!chain.equals(" "))
      {
        sb.append("." + chain);
      }
      String atomSpec = sb.toString();

      StringBuilder command = new StringBuilder(32);
      if (lastMousedOverAtomSpec != null)
      {
        command.append("~show " + lastMousedOverAtomSpec + ";");
      }
      viewerCommandHistory(false);
      command.append("show ").append(atomSpec);
      String cmd = command.toString();
      if (cmd.length() > 0)
      {
        viewer.sendChimeraCommand(cmd, false);
      }
      viewerCommandHistory(true);
      this.lastMousedOverAtomSpec = atomSpec;
    }
  }

  private void log(String message)
  {
    System.err.println("## Chimera log: " + message);
  }

  private void viewerCommandHistory(boolean enable)
  {
    // log("(Not yet implemented) History "
    // + ((debug || enable) ? "on" : "off"));
  }

  public void loadInline(String string)
  {
    loadedInline = true;
    // TODO: re JAL-623
    // viewer.loadInline(strModel, isAppend);
    // could do this:
    // construct fake fullPathName and fileName so we can identify the file
    // later.
    // Then, construct pass a reader for the string to Jmol.
    // ((org.jmol.Viewer.Viewer) viewer).loadModelFromFile(fullPathName,
    // fileName, null, reader, false, null, null, 0);
    // viewer.openStringInline(string);
    log("cannot load inline in Chimera, yet");
  }

  public void mouseOverStructure(int atomIndex, String strInfo)
  {
    // function to parse a mouseOver event from Chimera
    //
    int pdbResNum;
    int alocsep = strInfo.indexOf("^");
    int mdlSep = strInfo.indexOf("/");
    int chainSeparator = strInfo.indexOf(":"), chainSeparator1 = -1;

    if (chainSeparator == -1)
    {
      chainSeparator = strInfo.indexOf(".");
      if (mdlSep > -1 && mdlSep < chainSeparator)
      {
        chainSeparator1 = chainSeparator;
        chainSeparator = mdlSep;
      }
    }
    // handle insertion codes
    if (alocsep != -1)
    {
      pdbResNum = Integer.parseInt(strInfo.substring(
              strInfo.indexOf("]") + 1, alocsep));

    }
    else
    {
      pdbResNum = Integer.parseInt(strInfo.substring(
              strInfo.indexOf("]") + 1, chainSeparator));
    }
    String chainId;

    if (strInfo.indexOf(":") > -1)
    {
      chainId = strInfo.substring(strInfo.indexOf(":") + 1,
              strInfo.indexOf("."));
    }
    else
    {
      chainId = " ";
    }

    String pdbfilename = modelFileNames[frameNo]; // default is first or current
    // model
    if (mdlSep > -1)
    {
      if (chainSeparator1 == -1)
      {
        chainSeparator1 = strInfo.indexOf(".", mdlSep);
      }
      String mdlId = (chainSeparator1 > -1) ? strInfo.substring(mdlSep + 1,
              chainSeparator1) : strInfo.substring(mdlSep + 1);
      try
      {
        // recover PDB filename for the model hovered over.
        int _mp = _modelFileNameMap.length - 1, mnumber = new Integer(mdlId)
                .intValue() - 1;
        while (mnumber < _modelFileNameMap[_mp])
        {
          _mp--;
        }
        pdbfilename = modelFileNames[_mp];
        if (pdbfilename == null)
        {
          // pdbfilename = new File(viewer.getModelFileName(mnumber))
          // .getAbsolutePath();
        }

      } catch (Exception e)
      {
      }
      ;
    }
    if (lastMessage == null || !lastMessage.equals(strInfo))
    {
      getSsm().mouseOverStructure(pdbResNum, chainId, pdbfilename);
    }

    lastMessage = strInfo;
  }

  public void notifyAtomPicked(int atomIndex, String strInfo, String strData)
  {
    /**
     * this implements the toggle label behaviour copied from the original
     * structure viewer, MCView
     */
    if (strData != null)
    {
      System.err.println("Ignoring additional pick data string " + strData);
    }
    // rewrite these selections for chimera (DNA, RNA and protein)
    int chainSeparator = strInfo.indexOf(":");
    int p = 0;
    if (chainSeparator == -1)
    {
      chainSeparator = strInfo.indexOf(".");
    }

    String picked = strInfo.substring(strInfo.indexOf("]") + 1,
            chainSeparator);
    String mdlString = "";
    if ((p = strInfo.indexOf(":")) > -1)
    {
      picked += strInfo.substring(p + 1, strInfo.indexOf("."));
    }

    if ((p = strInfo.indexOf("/")) > -1)
    {
      mdlString += strInfo.substring(p, strInfo.indexOf(" #"));
    }
    picked = "((" + picked + ".CA" + mdlString + ")|(" + picked + ".P"
            + mdlString + "))";
    viewerCommandHistory(false);

    if (!atomsPicked.contains(picked))
    {
      viewer.select(picked);
      atomsPicked.add(picked);
    }
    else
    {
      viewer.select("not " + picked);
      atomsPicked.remove(picked);
    }
    viewerCommandHistory(true);
    // TODO: in application this happens
    //
    // if (scriptWindow != null)
    // {
    // scriptWindow.sendConsoleMessage(strInfo);
    // scriptWindow.sendConsoleMessage("\n");
    // }

  }

  // incremented every time a load notification is successfully handled -
  // lightweight mechanism for other threads to detect when they can start
  // referring to new structures.
  private long loadNotifiesHandled = 0;

  public long getLoadNotifiesHandled()
  {
    return loadNotifiesHandled;
  }

  public void notifyFileLoaded(String fullPathName, String fileName2,
          String modelName, String errorMsg, int modelParts)
  {
    if (errorMsg != null)
    {
      fileLoadingError = errorMsg;
      refreshGUI();
      return;
    }
    // TODO: deal sensibly with models loaded inLine:
    // modelName will be null, as will fullPathName.

    // the rest of this routine ignores the arguments, and simply interrogates
    // the Jmol view to find out what structures it contains, and adds them to
    // the structure selection manager.
    fileLoadingError = null;
    String[] oldmodels = modelFileNames;
    modelFileNames = null;
    chainNames = new ArrayList<String>();
    chainFile = new HashMap<String, String>();
    boolean notifyLoaded = false;
    String[] modelfilenames = getPdbFile();
    // first check if we've lost any structures
    if (oldmodels != null && oldmodels.length > 0)
    {
      int oldm = 0;
      for (int i = 0; i < oldmodels.length; i++)
      {
        for (int n = 0; n < modelfilenames.length; n++)
        {
          if (modelfilenames[n] == oldmodels[i])
          {
            oldmodels[i] = null;
            break;
          }
        }
        if (oldmodels[i] != null)
        {
          oldm++;
        }
      }
      if (oldm > 0)
      {
        String[] oldmfn = new String[oldm];
        oldm = 0;
        for (int i = 0; i < oldmodels.length; i++)
        {
          if (oldmodels[i] != null)
          {
            oldmfn[oldm++] = oldmodels[i];
          }
        }
        // deregister the Jmol instance for these structures - we'll add
        // ourselves again at the end for the current structure set.
        getSsm().removeStructureViewerListener(this, oldmfn);
      }
    }

    // register ourselves as a listener and notify the gui that it needs to
    // update itself.
    getSsm().addStructureViewerListener(this);

    if (notifyLoaded)
    {
      FeatureRenderer fr = getFeatureRenderer(null);
      if (fr != null)
      {
        fr.featuresAdded();
      }
      refreshGUI();
      loadNotifiesHandled++;
    }
    setLoadingFromArchive(false);
  }

  public void setJalviewColourScheme(ColourSchemeI cs)
  {
    colourBySequence = false;

    if (cs == null)
    {
      return;
    }

    int index;
    Color col;
    // Chimera expects RBG values in the range 0-1
    final double normalise = 255D;
    viewerCommandHistory(false);
    // TODO: Switch between nucleotide or aa selection expressions
    StringBuilder command = new StringBuilder(128);
    command.append("color white;");
    for (String res : ResidueProperties.aa3Hash.keySet())
    {
      index = ResidueProperties.aa3Hash.get(res).intValue();
      if (index > 20)
      {
        continue;
      }

      col = cs.findColour(ResidueProperties.aa[index].charAt(0));
      command.append("color " + col.getRed() / normalise + ","
              + col.getGreen() / normalise + "," + col.getBlue()
              / normalise + " ::" + res + ";");
    }

    evalStateCommand(command.toString(),false);
    viewerCommandHistory(true);
  }

  /**
   * called when the binding thinks the UI needs to be refreshed after a Chimera
   * state change. this could be because structures were loaded, or because an
   * error has occurred.
   */
  public abstract void refreshGUI();

  public void setLoadingFromArchive(boolean loadingFromArchive)
  {
    this.loadingFromArchive = loadingFromArchive;
  }

  /**
   * 
   * @return true if Chimeral is still restoring state or loading is still going
   *         on (see setFinsihedLoadingFromArchive)
   */
  public boolean isLoadingFromArchive()
  {
    return loadingFromArchive && !loadingFinished;
  }

  /**
   * modify flag which controls if sequence colouring events are honoured by the
   * binding. Should be true for normal operation
   * 
   * @param finishedLoading
   */
  public void setFinishedLoadingFromArchive(boolean finishedLoading)
  {
    loadingFinished = finishedLoading;
  }

  /**
   * Send the Chimera 'background solid <color>" command.
   * 
   * @see https
   *      ://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/background
   *      .html
   * @param col
   */
  public void setBackgroundColour(Color col)
  {
    viewerCommandHistory(false);
    double normalise = 255D;
    final String command = "background solid " + col.getRed() / normalise + ","
            + col.getGreen() / normalise + "," + col.getBlue()
            / normalise + ";";
    viewer.sendChimeraCommand(command, false);
    viewerCommandHistory(true);
  }

  /**
   * 
   * @param pdbfile
   * @return text report of alignment between pdbfile and any associated
   *         alignment sequences
   */
  public String printMapping(String pdbfile)
  {
    return getSsm().printMapping(pdbfile);
  }

  /**
   * Ask Chimera to save its session to the given file. Returns true if
   * successful, else false.
   * 
   * @param filepath
   * @return
   */
  public boolean saveSession(String filepath)
  {
    if (isChimeraRunning())
    {
      List<String> reply = viewer.sendChimeraCommand("save " + filepath,
              true);
      if (reply.contains("Session written"))
      {
        return true;
      }
      else
      {
        Cache.log
                .error("Error saving Chimera session: " + reply.toString());
      }
    }
    return false;
  }

  /**
   * Ask Chimera to open a session file. Returns true if successful, else false.
   * The filename must have a .py extension for this command to work.
   * 
   * @param filepath
   * @return
   */
  public boolean openSession(String filepath)
  {
    evalStateCommand("open " + filepath, true);
    // todo: test for failure - how?
    return true;
  }

  public boolean isFinishedInit()
  {
    return finishedInit;
  }

  public void setFinishedInit(boolean finishedInit)
  {
    this.finishedInit = finishedInit;
  }

  public List<String> getChainNames()
  {
    return chainNames;
  }

}
